from typing import List, Optional
from pydantic import BaseModel
import yaml
import os
from .context import Context

class Stack(BaseModel):
    path: str
    name: str
    parameters: Optional[dict]

    @property
    def _path(self):
        return self.path.replace("$root", Context.get_root())
        
    @property
    def _yml(self):
        class SafeLoaderIgnoreUnknown(yaml.SafeLoader):
            def ignore_unknown(self, node):
                return None 

        SafeLoaderIgnoreUnknown.add_constructor(None, SafeLoaderIgnoreUnknown.ignore_unknown)

        return yaml.load(open(self._path).read(), Loader=SafeLoaderIgnoreUnknown)

    @property
    def resources(self):
        return self._yml.get("Resources", {})

class Enviroment(BaseModel):
    artifacts: str 
    name: str
    region: Optional[str] = None
    profile: Optional[str] = None



class Config(BaseModel):
    Stacks: List[Stack]
    Enviroments: List[Enviroment]

    @staticmethod
    def parse(path: str = None):
        try:
            data = yaml.safe_load(open(path))
        except:
            raise IOError("Not able to find file at path: " + '"' + path + '"')

        try:
            return Config(**data)
        except Exception as e:
            raise Exception("Not able to pase config: " + '"' + str(e) + '"')

    def setup_env(self, env=None):
        if not env:
            if self.Enviroments[0].profile:
                os.environ["AWS_PROFILE"] = self.Enviroments[0].profile

            if self.Enviroments[0].region:
                os.environ["AWS_DEFAULT_REGION"] = self.Enviroments[0].region



import argparse

from .meta import VERSION
from .commands.deploy import deploy
from .commands.diff import diff
from .commands.info import info
from .utils.logging import logger
from .utils.context import Context

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "action", 
        choices=['diff', 'info', 'deploy', 'package', "version"], 
        help='what action to preform'
    )
    parser.add_argument(
        "-path", "--path",
        default="services.yml",
        help='path to the file describing the services'
    )

    parser.add_argument("-v", '--version', action='version', help='path to the file describing the services', version='aws-cf: ' + VERSION)
    parser.add_argument("-r", "--root")
    parser.add_argument("-vb", "--verbose", type=bool)
    

    args = parser.parse_args()

    try:
        Context.set_root(args.root or ".")
        Context.set_service_path(args.path)

        if args.action == "deploy":
            deploy(args.path, args.root or ".")

        if args.action == "info":
            info()

        if args.action == "diff":
            diff(args.path, args.root or ".")

    except Exception as e:
        if args.verbose:
            raise e
        logger.error(str(e))



if __name__ == '__main__':
    main()


from ..utils.logging import logger
from ..utils.config import Config
from ..utils.context import Context

def info():
    config = Config.parse(Context.get_service_path())
  
    envs = '\n'.join(["  🌳 " + env.name + " (" + env.region + ")"  for env in config.Enviroments])
    stacks = '\n'.join([f"  📚 {stack.name} ({len(stack.resources.keys())} resource)" for stack in config.Stacks])
    root_path = Context.get_root()
    service_path = Context.get_service_path()

    logger.warn(f"""

Meta:
  🛎️  services: {service_path}
  🪵  root directory: {root_path}

Enviroments: 
{envs} 

Stacks:
{stacks}
    """)


from ..utils.logging import logger
from ..utils.config import Config
import sys
from ..utils.common import create_change_set, get_yml, remove_change_set, format_diff


def diff(config_path, root_path):
    config = Config.parse(config_path)
    config.setup_env()

    services = config.Stacks    

    for service in services:
        change_set = create_change_set(service, config)

        if change_set:
            diffs = [format_diff(change)for change in change_set["Changes"]]
            logger.warn(f"{service.name} (changes {len(diffs)})")

            if len(diffs):
                logger.warning(f"  🆕  Found {len(diffs)} differences for the stack {service.name}")
                for diff in diffs:
                    logger.warning(f"> {diff}")
            else:
                logger.info(f"  No changes")

            remove_change_set(service.name, change_set["ChangeSetName"])
        
        else:
            yml = get_yml(service.path, config, root_path)
            logger.warn(f"{service.name} new stack ⭐")
            logger.warn(yml)


import boto3
import botocore.exceptions
import time
import subprocess
from .config import Config, Stack
from .context import Context
import tempfile
import json

def get_yml(path, config, root_path):
    path = path.replace("$root", root_path)
    return package(open(path).read(), config)

def create_change_set(stack: Stack, config: Config):
    PREFIX = Context.get_changeset_prefix()

    root_path = Context.get_root()

    path = stack.path
    path = path.replace("$root", root_path)
    client = boto3.client("cloudformation")
    name = stack.name

    try:
        previouse_change_sets = client.list_change_sets(StackName=stack.name)
    except botocore.exceptions.ClientError as e:
        if str(e).endswith(f"Stack [{stack.name}] does not exist"):
            return None
        raise e

    def get_name(previouse_change_sets):
        if not len(previouse_change_sets["Summaries"]):
            return PREFIX + "-" + name + "-" + str(1)

        previouse_name:str = previouse_change_sets["Summaries"][0]["ChangeSetName"]

        if not previouse_name.startswith(PREFIX):
            return PREFIX + "-" + name + "-" + str(1)

        previouse_index = int(previouse_name.split("-")[2])
        new_index = previouse_index + 1
        return PREFIX + name + "-" + str(new_index)

    change_set_name = get_name(previouse_change_sets)
    parameters  =[]
    
    if stack.parameters:
        parameters = [{"ParameterKey": key, "ParameterValue": stack.parameters[key]} for key in stack.parameters.keys()]

    client.create_change_set(
        ChangeSetName=change_set_name,
        StackName=name,
        Capabilities=["CAPABILITY_NAMED_IAM"],
        TemplateBody=get_yml(path, config, root_path),
        Parameters=parameters,
    )
    wait_for_ready(name, change_set_name)

    return client.describe_change_set(
        ChangeSetName=change_set_name,
        StackName=name
    )


def wait_for_ready(name, change_set_name):
    client = boto3.client("cloudformation") 
    while True:
        response = client.describe_change_set(
            ChangeSetName=change_set_name,
            StackName=name,
        )
    
        response = client.describe_change_set(
            ChangeSetName=change_set_name,
            StackName=name,
        )

        if response["Status"] not in ["CREATE_PENDING", "CREATE_IN_PROGRESS"]:
            break

        time.sleep(3)

def remove_change_set(name: str, change_set_name: str):
    client = boto3.client("cloudformation")

    response = client.delete_change_set(
        ChangeSetName=change_set_name,
        StackName=name
    )

def format_diff(diff):
    action = diff["ResourceChange"]["Action"]
    resource_id = diff["ResourceChange"]["LogicalResourceId"]
    resource_type = diff["ResourceChange"]["ResourceType"]
    details = diff["ResourceChange"]["Details"]

    actionName = {
        "Add": "Adding",
        "Modify": "Modifying",
        "Remove": "Removing"
    }

    if len(details):
        return f"{actionName[action]} {resource_type} with id {resource_id} \n{json.dumps(details)}\n\n"
        
    return f"{actionName[action]} {resource_type} with id {resource_id}"

def deploy_stack(name: str, change_set):
    client = boto3.client("cloudformation")
    response = client.execute_change_set(
        ChangeSetName=change_set,
        StackName=name
    )

def create_stack(name: str, template:str):
    client = boto3.client("cloudformation")
    response = client.create_stack(
        StackName=name,
        TemplateBody=template
    )

def package(yml: str, config: Config):
    with tempfile.NamedTemporaryFile() as tmp:
        tmp.write(bytes(yml, "utf-8"))
        tmp.read()
    
        args = [
                "aws", "cloudformation", "package",
                "--template", tmp.name,
                "--s3-prefix", "aws/stacks",
                "--s3-bucket", config.Enviroments[0].artifacts,
        ]
        
        if config.Enviroments[0].profile:
            args.append("--profile")
            args.append(config.Enviroments[0].profile)

        result = subprocess.check_output(args)
        
        return result.decode()

def get_yes_or_no(message):
    while True:
        result = input(message + " (enter y/n)")

        if result in ["yes", "y"]:
            return True

        if result in ["no", "n"]:
            return True